*
* "KnightKit" source code.
* QL Utility toolkit, by M. Knight.
* Written 1994, 1995. Public domain.
*
* Stage one; SYSVARS, a function to return address of system variables,
*            and CHVARS, a function to return address of a channel's variables.
*            CHVARS rejects most non-screen channels, returns -15 for
*            "Bad parameter" instead.
*
* Stage two; GRAB_MEMORY, a function to return base address of reserved area or
*            error code if anything goes wrong.  RELEASE_MEMORY, to return the
*            area to Q-DOS, and FREEMEM, no parameters, returns free memory
*            in the system.
*
* Stage three; QDOS$, Q-DOS version returned as a string of four characters.
*            This function complements VER$.
*
Initialise LEA.L       Define,A1       Move address of definitions to A1.
           MOVE.W      $110,A2         Place vector in A2.
           JMP         (A2)            Jump to QL ROM routine - it returns.
*
* AnyVariable=GRAB_MEMORY(NumBytes)
*
GrabMem    MOVE.W      $118,A2         Vector to fetch long integer parameters.
           JSR         (A2)            Fetch one...
           BNE         Bad_Exit        If there were none, report error.
           SUBQ.W      #1,D3           Was there just one?
           BEQ.S       BytesFound      If one, then ok.
           BRA         Bad_Param       More or less than one parameter found.
BytesFound MOVE.L      0(A1,A6.L),D1   Move the number of bytes into D1.
           BLE         Bad_Param       If less than one byte, reject it.
           ADDQ.L      #4,D1           Add 4 to number of bytes, for marker.
           MOVE.L      #-1,D2          This job will own the memory.
           ADDQ.L      #4,$58(A6)      Tidy maths stack.
           MOVEQ       #$18,D0         Select MT.ALCHP, grab some heap memory.
           TRAP        #1              Call it to reserve the memory.
           TST.L       D0              Was there an error?
           BMI.S       GrabError       If yes, return the error code.
           MOVE.L      #'GRAB',(A0)    Put marker in the area reserved.
           ADDA.L      #4,A0           Add 4 to base, avoid overwriting marker.
           MOVE.L      A0,D1           Move the address into D1.
           BRA         D1_Ret          Return the address to BASIC.
GrabError  MOVE.L      D0,D1           Move the error code to D1.
           BRA         D1_Ret          Return the error code to BASIC.
*
* RELEASE_MEMORY Address   Where "Address" is a base address of a previously
*                          reserved area.
*
Release    MOVE.W      $118,A2         Vector to fetch long integer parameters.
           JSR         (A2)            Fetch one...
           BNE.S       Bad_Exit        If there were none, report error.
           SUBQ.W      #1,D3           Was there just one?
           BEQ.S       AddrFound       If one, then ok.
           BRA.S       Bad_Param       More or less than one parameter found.
AddrFound  MOVE.L      0(A1,A6.L),D1   Fetch the parameter from stack.
           BMI.S       Bad_Param       Reject if it is negative.
           BTST        #0,D1           Is it odd?
           BNE.S       Bad_Param       Reject if it is odd.
           MOVE.L      D1,A0           Move address to A0 to use as pointer.
           SUBA.L      #4,A0           Subtract 4 to look for marker.
           CMPI.L      #'GRAB',(A0)    Is it ours?
           BNE.S       Bad_Param       It isn't our memory, so report error.
           CLR.L       (A0)            Remove the marker.
           MOVEQ       #$19,D0         Select MT.RECHP, return the memory.
           TRAP        #1              Call the system routine to return it.
           MOVEQ       #0,D0           Signal no error.
           RTS                         Return to BASIC.
*
* AnyVariable=CHVARS(#Ch) '#' is optional, Ch is number of an open channel.
*
ChVars     MOVE.W      $112,A2         Vector to fetch integer parameters.
           JSR         (A2)            Fetch it...
           BNE.S       Bad_Exit        If there were none, report the error.
           SUBQ.W      #1,D3           Was there just one?
           BEQ.S       Got_It          OK, there was just one.
           BRA.S       Bad_Param       There was no parameter or more than one.
Got_It     MOVE.W      0(A1,A6.L),D0   Put the channel number in D0.
           ADDQ.L      #2,$58(A6)      Tidy maths stack.
Chan_Sel   MULU        #40,D0          Channel table size is 40 bytes.
           ADD.L       $30(A6),D0      Add offset for base address.
           CMP.L       $34(A6),D0      Are we past end of table?
           BGE.S       What_Chan       Past end of table, so not open.
           MOVE.L      0(A6,D0.L),D0   Fetch channel ID.
           BMI.S       What_Chan       If negative then the channel is closed.
Got_Id     MOVE.L      D0,A0           Now A0 is channel ID.
           LEA.L       ExTopper,A2     A2 now points to extension code.
           MOVEQ       #0,D3           Timeout is zero, return immediately.
           MOVEQ       #9,D0           Select SD.EXTOP routine...
           TRAP        #3              ...and then call it.
           TST.L       D0              Any error?
           BEQ.S       D1_Ret          If not, result is in D1, so return it.
           MOVE.L      D0,D1           Move error code to D1 if there was one.
           BRA.S       D1_Ret          Return the error code instead.
What_Chan  MOVEQ       #-6,D0          Return with "Channel not open" error.
           RTS
Bad_Param  MOVEQ       #-15,D0         Return with "Bad parameter" error.
Bad_Exit   RTS
ExTopper   MOVE.L      A0,D1           Return the channel base.
           MOVEQ       #0,D0           No error if we have got this far.
           RTS
*
* SYSVARS  Function, no parameters, returns address of system variables.
*
SysVars    MOVEQ       #0,D0           Select MT.INF, gives system information.
           TRAP        #1              Call routine, leaves result in A0.L
           MOVE.L      A0,D1           Move longword address to D1.L
*
* How to return a LONGWORD value, coerced into FLOATING POINT,
* To SuperBASIC from a FUNCTION extension.
*
* At the start of the code, the LONGWORD value is in data register D1.
*
D1_Ret     MOVE.W      D1,D4           D4 will be exponent.
           MOVE.L      D1,D5           D5 will be mantissa.
           BEQ.S       Norlised        Zero is a trivial case.
           MOVE.W      #2079,D4        First guess at exponent.
           ADD.L       D1,D1           Is it normalised?
           BVS.S       Norlised        If so, return it.
           SUBQ.W      #1,D4           Not ready, halve exponent.
           MOVE.L      D1,D5           Doubled mantissa moved to D5.
           MOVEQ       #16,D0          Try 16 bit shift first.
*
Normalise  MOVE.L      D5,D1           Make copy of mantissa.
           ASL.L       D0,D1           Shift mantissa D0 places left.
           BVS.S       Too_Far         Try next shift distance.
           SUB.W       D0,D4           correct exponent for that shift.
           MOVE.L      D1,D5           New mantissa is closer.
Too_Far    ASR.W       #1,D0           Half the shift distance for next try.
           BNE.S       Normalise       Half the previous shift each time.
*
* Make sure there is enough space for the result.
*
Norlised   MOVEQ       #6,D1           Number of bytes needed is 6.
           MOVE.W      $11A,A0         Select BV.CHRIX to reserve space.
           JSR         (A0)            Call the routine
           MOVE.L      $58(A6),A1      Fetch safe A1 value.
           SUBQ.L      #6,A1           Point to the new value.
           MOVE.L      A1,$58(A6)      Update stack pointer.
           MOVE.L      D5,2(A1,A6.L)   Stack the mantissa.
           MOVE.W      D4,0(A1,A6.L)   Stack the exponent.
           MOVEQ       #2,D4           Indicate floating point result.
           MOVEQ       #0,D0           Indicate no error.
           RTS
*
* AnyVariable=FREEMEM  No parameters required.
*
FreeMem    MOVEQ       #0,D0           Select MT.INF, gives system information.
           TRAP        #1              Call routine, leaves result in A0.L
           MOVE.L      12(A0),A2       Move lowmem pointer into A2.
           MOVE.L      16(A0),A3       Move highmem pointer into A3.
           SUBA.L      A2,A3           Subtract lowmem from highmem.
           MOVE.L      A3,D1           Move free memory figure into D1.
           BRA.S       D1_Ret          Return the result to BASIC.
*
*          QDOS$ Extension to SuperBASIC
*          as written by M Knight.
*
DosVer     MOVEQ       #0,D0           Select MT.INF ROM routine.
           TRAP        #1              Call it.
           MOVE.L      D2,D7           Save the QDOS version in D7.
           MOVEQ       #6,D1           Indicate 6 bytes required on stack.
           MOVE.W      $11A,A2         Select BV.CHRIX, reserve stack space...
           JSR         (A2)            ...and call the routine.
           MOVE.L      $58(A6),A1      Find stack pointer.
           SUBQ.L      #6,A1           Make room on the stack.
           MOVE.L      D7,2(A1,A6.L)   Stack the version number.
           MOVE.W      #4,0(A1,A6.L)   Stack the string length.
           MOVE.L      A1,$58(A6)      Save the new stack pointer.
           MOVEQ       #1,D4           Indicate string result.
           MOVEQ       #0,D0           Indicate no error.
           RTS                         Return the result.
*
Define     DC.W        2               One PROCedure, but with a long name.
           DC.W        Release-*       Point to code for RELEASE_MEMORY.
           DC.B        14,'RELEASE_MEMORY' Length of name, then name.
           DC.W        0               End of PROCedure list.
           DC.W        6               Five FuNctions, but long name for one.
           DC.W        SysVars-*       Pointer to SYSVARS code.
           DC.B        7,'SYSVARS'     Length of name, then characters of name.
           DC.W        ChVars-*        Pointer to CHVARS code.
           DC.B        6,'CHVARS'      Length of name, then characters of name.
           DC.W        GrabMem-*       Point to code for GRAB_MEMORY.
           DC.B        11,'GRAB_MEMORY' Length of name, then characters of name.
           DC.W        FreeMem-*       Pointer to FREEMEM code.
           DC.B        7,'FREEMEM'     Length of name, then characters of name.
           DC.W        DosVer-*        Pointer to QDOS$ code.
           DC.B        5,'QDOS$'       Length of name, then characters of name.
           DC.W        0               End of function list.
           END
